@@ -462,23 +462,27 @@ displayChangeBarrier :: TVar Barrier
 displayChangeBarrier = unsafePerformIO $ newTVarIO 0
 
 -- | Runs a STM action, and waits for the display to be fully updated
--- before returning.
+-- with any changes that action makes to the displayed regions.
 waitDisplayChange :: STM a -> IO a
 waitDisplayChange a = do
-	r <- atomically a
 	c <- atomically $ dupTChan displayUpdateNotifier
-	b <- atomically $ do
-		!b <- succ <$> readTVar displayChangeBarrier
-		writeTVar displayChangeBarrier b
-		return b
-	atomically $ waitchange c b
-	return r
+	bv <- newEmptyTMVarIO
+	setbarrier bv `concurrently` waitchange c bv
+	snd <$> atomically (readTMVar bv)
   where
-	waitchange c b = do
-		change <- readTChan c
+	setbarrier bv = atomically $ do
+		!b <- succ <$> readTVar displayChangeBarrier
+		r <- a
+		writeTVar displayChangeBarrier b
+		putTMVar bv (b, r)
+	waitchange c bv = do
+		change <- atomically $ readTChan c
+		-- this blocks until the STM action has run, and the
+		-- barrier is set.
+		b <- fst <$> atomically (readTMVar bv)
 		case change of
-			DisplayChangeBarrier b' | b' == b -> return ()
-			_ -> waitchange c b
+			DisplayChangeBarrier b' | b' >= b -> return ()
+			_ -> waitchange c bv
 
 displayThread :: Bool -> TSem -> IO ()
 displayThread isterm endsignal = do
